/*	AmpEnvRelease.cpp    */

/*
	The AM1 Synthesizer for Macintosh, version 0.50 beta
	Copyright (C) 2003 The University of Toledo.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

	1.    Redistributions of source code must retain the above copyright notice,
	      this list of conditions and the following disclaimer. 
	2.    Redistributions in binary form must reproduce the above copyright notice,
	      this list of conditions and the following disclaimer in the documentation
	      and/or other materials provided with the distribution. 
	3.    The names of the authors and of the University of Toledo may not be used
	      to endorse or promote products derived from this software without specific
	      prior written permission. 

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND THE UNIVERSITY OF TOLEDO "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR THE UNIVERSITY OF TOLEDO BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include	"AmpEnvRelease.h"
#include	<cstdlib>
#include	<climits>

using	namespace	std;
using	namespace	AnthonyAudioFramework;

namespace	AnthonyAudioClasses	{

AmpEnvRelease::AmpEnvRelease( AudioAbstract* out, size_t port, vector< std::pair<double,double> > stages,
					size_t sustain, size_t release ) : AudioAbstract( out, port )
{
	Init();
	SetStages( stages, sustain, release );
}

AmpEnvRelease::~AmpEnvRelease()
{
	if	(_stageDurs)	delete [] _stageDurs;
	if	(_stageAmps)	delete [] _stageAmps;
	if	(_stageCounts)	delete [] _stageCounts;
}

size_t AmpEnvRelease::GetSamples( AudioSample* buffer, size_t samplesWanted )
{
	size_t	samplesReturned;
	
	samplesReturned = Previous(0)->GetSamples( buffer, samplesWanted );
	
	// modify amplitude of samples
	double		deltaAmp = _stageAmps[_currentStage] - _currentAmp;
	double		stageSamplesLeft = (double)(_stageCounts[_currentStage] - _samplecount);
	
	for	( int sample = 0; sample < samplesReturned; sample++ )	{
		if	( !_atEnd )	{
			if	( !_isSustaining and _samplecount != 0 )							// modify amp if not sustaining
				_currentAmp += deltaAmp / stageSamplesLeft;
			buffer[sample] = (AudioSample)( _currentAmp * (double)buffer[sample] );			// scale sample
			
			_samplecount++;
			if	( _samplecount >= _stageCounts[_currentStage] )	{					// if at end of current stage, and
				if	( !_isReleasing and _currentStage == _sustainStage )	{			// if sustain stage
					_isSustaining = true;									// then sustain
					_currentAmp = _stageAmps[_currentStage];
				}
				else	{												// if not sustain stage
					_currentStage++;										// begin next stage
					if	( _currentStage < _numStages )	{
						_currentAmp = _stageAmps[_currentStage-1];				// (trying to avoid accumulation of FP errors)
						deltaAmp = _stageAmps[_currentStage] - _currentAmp;
						stageSamplesLeft = (double)_stageCounts[_currentStage];
						_samplecount = 0;
					}
					else	{											// past the last stage
						_atEnd = true;
						_isSustaining = false;
						_isReleasing = false;
					}
				}
			}
		}
		else	buffer[sample] = 0;											// pin output at zero at end of envelope
	}
	
	return samplesReturned;	 
}

void	AmpEnvRelease::SetStages( vector< std::pair<double,double> > &stages, size_t sustain, size_t release )
{
	// delete old arrays if they exist
	if	(_stageDurs)	delete [] _stageDurs;
	if	(_stageAmps)	delete [] _stageAmps;
	if	(_stageCounts)	delete [] _stageCounts;
	
	// create arrays to store stage parameters
	_numStages = stages.size();
	_stageDurs = new double[_numStages];
	_stageAmps = new double[_numStages];
	_stageCounts = new unsigned long[_numStages];
	if	( !_stageDurs or !_stageAmps or !_stageCounts )	{
		cerr	<< "Could not allocate enough memory for an AmpEnvRelease.\n";
		exit(1);
	}
	
	// copy values from vector into arrays
	vector< pair<double,double> >::iterator	iter = stages.begin();
	for	( int i = 0; (iter != stages.end()) && (i < _numStages); ++iter, ++i )	{
		_stageDurs[i] = iter->first;
		_stageAmps[i] = iter->second;
	}
	
	// set the sustain and release stages
	_sustainStage = sustain;
	_releaseStage = release;
	
	return;
}

void	AmpEnvRelease::TriggerRelease()
{
	if	( !_isReleasing and !_atEnd )	{
		_isSustaining = false;
		_isReleasing = true;
		_currentStage = _releaseStage;
		_samplecount = 0;
	}
	
	return;
}

void	AmpEnvRelease::Init()
{
	_numStages = 0;
	_currentStage = 0;
	_sustainStage = 0;
	_releaseStage = 0;
	_isSustaining = false;
	_isReleasing = false;
	_atEnd = false;
	_stageDurs = NULL;
	_stageAmps = NULL;
	_stageCounts = NULL;
	_currentAmp = 0.0;
	_samplecount = 0;
	
	return;
}

void	AmpEnvRelease::Reset()
{
	AudioAbstract::Reset();
	ResetState();
	return;
}

void	AmpEnvRelease::InitializeForPlay()
{
	AudioAbstract::InitializeForPlay();
	ResetState();
	return;
}

void	AmpEnvRelease::ResetState()
{
	double srate = (double)( this->SamplingRate() );
	double durSamples = 0.0;

	// convert stage durations into ending sample counts
	for	( int i = 0; i < _numStages; i++ )	{
		durSamples = _stageDurs[i] * srate;
		if	( _stageCounts[i] <= (double) ULONG_MAX )	{
			_stageCounts[i] = (unsigned long)durSamples;
		}
		else	{
			cerr	<< "Duration exceeded limits of an amplitude envelope object.\n";
			exit(1);
		}
	}	
	
	_currentStage = 0;
	_isSustaining = false;
	_isReleasing = false;
	_atEnd = false;
	_currentAmp = 0.0;
	_samplecount = 0;
	
	return;
}

}	// namespace closure
