/*	AmpEnvRelease.h

	An audio rate envelope which modifies the amplitude of its input.
	Capable of an arbitrary number of stages with linear transitions (no looping).
	May mark one stage each as the "sustain" and "release" stages so that 
	Midi notes may be played in real-time with proper Note Off behavior.
	Also differs from the class AmpEnvelope in that stage times are absolute
	times in seconds instead of percentages of the total duration.
	
	Anthony Kozar
	March 5, 2003
 */

/*
	The AM1 Synthesizer for Macintosh, version 0.50 beta
	Copyright (C) 2003 The University of Toledo.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

	1.    Redistributions of source code must retain the above copyright notice,
	      this list of conditions and the following disclaimer. 
	2.    Redistributions in binary form must reproduce the above copyright notice,
	      this list of conditions and the following disclaimer in the documentation
	      and/or other materials provided with the distribution. 
	3.    The names of the authors and of the University of Toledo may not be used
	      to endorse or promote products derived from this software without specific
	      prior written permission. 

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND THE UNIVERSITY OF TOLEDO "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR THE UNIVERSITY OF TOLEDO BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef	ANTHONY_AMP_ENV_RELEASE_H
#define	ANTHONY_AMP_ENV_RELEASE_H

#include	<vector>
#include	<utility>
#include	<cstdlib>
#include	"AudioAbstract.h"

namespace AAF = AnthonyAudioFramework;

namespace	AnthonyAudioClasses	{

class	AmpEnvRelease : public AAF::AudioAbstract
{
	public:
		AmpEnvRelease() : AAF::AudioAbstract() { Init(); }
		AmpEnvRelease( AAF::AudioAbstract* out, std::size_t port ) : AAF::AudioAbstract( out, port ) { Init(); }
		AmpEnvRelease( AAF::AudioAbstract* out, std::size_t port, std::vector< std::pair<double,double> > stages,
					std::size_t sustain, std::size_t release );
		virtual ~AmpEnvRelease();
		
		virtual std::size_t	PreviousCount()	{ return (std::size_t)1; }
		virtual std::size_t	GetSamples( AAF::AudioSample* buffer, std::size_t n );
		virtual void		Reset();
				
		// Each pair is a duration and an amplitude for a stage of the envelope.
		// Amplitude values are between 0.0 and 1.0. Durations are in seconds.
		// sustain is the number of the stage to sustain on and release is the 
		// number of the stage to jump to when TriggerRelease() is called.
		void	SetStages( std::vector< std::pair<double,double> > &stages, std::size_t sustain, std::size_t release );
		
		// Call TriggerRelease() when the Midi Note Off message is received.
		void	TriggerRelease();
		
		// These functions return the current state of the envelope
		bool	isSustaining()	{ return _isSustaining; }
		bool	isReleasing()	{ return _isReleasing; }

	protected:
		virtual void	InitializeForPlay();

	private:
		std::size_t		_numStages;
		std::size_t		_currentStage;
		std::size_t		_sustainStage;
		std::size_t		_releaseStage;
		bool			_isSustaining;
		bool			_isReleasing;
		bool			_atEnd;
		double*		_stageDurs;
		double*		_stageAmps;
		double		_currentAmp;
		unsigned long	_samplecount;							// number of samples since the beginning of current stage
		unsigned long*	_stageCounts;							// stage endpoints in samples
				
		void	Init();
		void	ResetState();
};

}	// namespace closure

#endif
