/*	AmpEnvelope.cpp    */

/*
	The AM1 Synthesizer for Macintosh, version 0.50 beta
	Copyright (C) 2003 The University of Toledo.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

	1.    Redistributions of source code must retain the above copyright notice,
	      this list of conditions and the following disclaimer. 
	2.    Redistributions in binary form must reproduce the above copyright notice,
	      this list of conditions and the following disclaimer in the documentation
	      and/or other materials provided with the distribution. 
	3.    The names of the authors and of the University of Toledo may not be used
	      to endorse or promote products derived from this software without specific
	      prior written permission. 

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND THE UNIVERSITY OF TOLEDO "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR THE UNIVERSITY OF TOLEDO BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include	"AmpEnvelope.h"
#include	<cstdlib>
#include	<climits>

using	namespace	std;
using	namespace	AnthonyAudioFramework;

namespace	AnthonyAudioClasses	{

AmpEnvelope::AmpEnvelope( AudioAbstract* out, std::size_t port, double dur, vector< pair<double,double> > stages )
	: AudioAbstract( out, port )
{
	Init();
	_duration = dur;
	SetStages(stages);
}

AmpEnvelope::~AmpEnvelope()
{
	if	(_stageDurs)	delete [] _stageDurs;
	if	(_stageAmps)	delete [] _stageAmps;
	if	(_stageTimes)	delete [] _stageTimes;
}

size_t AmpEnvelope::GetSamples( AudioSample* buffer, size_t samplesWanted )
{
	size_t	samplesReturned;
	
	samplesReturned = Previous(0)->GetSamples( buffer, samplesWanted );
	
	// modify amplitude of samples
	double		amp;
	double		deltaAmp = _stageAmps[_currentStage+1] - _stageAmps[_currentStage];
	double		stageLength = (double)(_stageTimes[_currentStage+1] - _stageTimes[_currentStage]);
	unsigned int	offset;
	
	for	( int sample = 0; sample < samplesReturned; sample++ )	{
		if	( _currentStage < _numStages )	{
			offset = _samplecount - _stageTimes[_currentStage];
			amp = _stageAmps[_currentStage] + deltaAmp * ( (double)offset / stageLength );
			buffer[sample] = (AudioSample)( amp * (double)buffer[sample] );
			
			_samplecount++;
			if	( _samplecount >= _stageTimes[_currentStage+1] )	{
				if	( ++_currentStage < _numStages )	{
					deltaAmp = _stageAmps[_currentStage+1] - _stageAmps[_currentStage];
					stageLength = (double)(_stageTimes[_currentStage+1] - _stageTimes[_currentStage]);
				}
			}
		}
		else	buffer[sample] = 0;							// pin output at zero at end of envelope
	}
	
	return samplesReturned;	 
}

void	AmpEnvelope::SetStages( vector< pair<double,double> > &stages )
{
	// delete old arrays if they exist
	if	(_stageDurs)	delete [] _stageDurs;
	if	(_stageAmps)	delete [] _stageAmps;
	if	(_stageTimes)	delete [] _stageTimes;
	
	// create arrays to store stage parameters
	_numStages = stages.size();
	_stageDurs = new double[_numStages+1];								// make space for implied first point
	_stageAmps = new double[_numStages+1];
	_stageTimes = new unsigned long[_numStages+1];
	if	( !_stageDurs or !_stageAmps or !_stageTimes )	{
		cerr	<< "Could not allocate enough memory for an AmpEnvelope.\n";
		exit(1);
	}
	
	// first point of envelope is always (0,0)
	_stageDurs[0] = 0.0;
	_stageAmps[0] = 0.0;
	
	// copy values from vector into arrays
	vector< pair<double,double> >::iterator	iter = stages.begin();
	for	( int i = 1; (iter != stages.end()) && (i <= _numStages); ++iter, ++i )	{
		_stageDurs[i] = iter->first;
		_stageAmps[i] = iter->second;
	}
	
	return;
}

void	AmpEnvelope::Init()
{
	_numStages = 0;
	_currentStage = 0;
	_stageDurs = NULL;
	_stageAmps = NULL;
	_stageTimes = NULL;
	_duration = 0.0;
	_samplecount = 0;
	
	return;
}

void	AmpEnvelope::Reset()
{
	AudioAbstract::Reset();
	ResetState();
	return;
}

void	AmpEnvelope::InitializeForPlay()
{
	AudioAbstract::InitializeForPlay();
	ResetState();
	return;
}

void	AmpEnvelope::ResetState()
{
	double srate = (double)( this->SamplingRate() );
	double durSamples = (_duration * srate);
	double durTotal;
	
	if	( durSamples <= (double) ULONG_MAX )	{
		// convert stage durations into ending sample counts
		_stageTimes[0] = 0;
		durTotal = 0.0;
		for	( int i = 1; i <= _numStages; i++ )	{
			durTotal += _stageDurs[i];
			_stageTimes[i] = (unsigned long)(durTotal * _duration * srate);
		}
	}	
	else	{
		cerr	<< "Duration exceeded limits of an amplitude envelope object.\n";
		exit(1);
	}
	
	_currentStage = 0;
	_samplecount = 0;
	
	return;
}

}	// namespace closure
