/*	AbstractMixer.cpp    */
 
/*
	The AM1 Synthesizer for Macintosh, version 0.50 beta
	Copyright (C) 2003 The University of Toledo.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

	1.    Redistributions of source code must retain the above copyright notice,
	      this list of conditions and the following disclaimer. 
	2.    Redistributions in binary form must reproduce the above copyright notice,
	      this list of conditions and the following disclaimer in the documentation
	      and/or other materials provided with the distribution. 
	3.    The names of the authors and of the University of Toledo may not be used
	      to endorse or promote products derived from this software without specific
	      prior written permission. 

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND THE UNIVERSITY OF TOLEDO "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR THE UNIVERSITY OF TOLEDO BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include	<cstdlib>
#include	<cassert>
#include	"AbstractMixer.h"

using	namespace	AnthonyAudioFramework;

namespace	AnthonyAudioClasses	{

AbstractMixer::AbstractMixer( AAF::AudioAbstract* out, size_t port, size_t numInputs ) : AAF::AudioAbstract( out, port )
{
	Init( numInputs );
}

AbstractMixer::~AbstractMixer()
{
	DeleteInputBuffers();
}

void	AbstractMixer::SetNumInputs( size_t n )
{
	DeleteInputBuffers();										// use old _numInputs when deleting
	_numInputs = n;

	if	( n > 0 )	{
		//AllocateInputBuffers();								// use new value when allocating
		AllocatePrevious();
	}
	
	return;
}

void	AbstractMixer::Init( size_t numInputs )
{ 
	_numInputs = 0;
	_inputBuffers = NULL;
	SetNumInputs( numInputs );
	
	return;
}

void	AbstractMixer::InitializeForPlay()
{
	AudioAbstract::InitializeForPlay();								// must call superclass method first
	AllocateInputBuffers();
	
	return;
}

void	AbstractMixer::DeleteInputBuffers()
{
	if	(_inputBuffers)	{
		for	( int i = 0; i < _numInputs; i++ )
			delete [] _inputBuffers[i];
		delete [] _inputBuffers;
	}
	_inputBuffers = NULL;
	
	return;
}

void	AbstractMixer::AllocateInputBuffers()
{
	int	numChannels = Channels();
	
	_inputBuffers = new AudioSample* [_numInputs];
	assert(_inputBuffers);
	for	( int i = 0; i < _numInputs; i++ )	{
		_inputBuffers[i] = new AudioSample[SAMPLES_PER_FRAME * numChannels];
		assert( _inputBuffers[i] );
	}
	
	return;
}

size_t AbstractMixer::GetSamples( AudioSample* buffer, size_t samplesWanted )
{
	size_t	samplesRead, samplesWritten;
	int		channels;
	bool		stop;
	
	stop = false;
	channels = Channels();
	samplesWritten = 0;
	
	while	( (samplesWanted - samplesWritten) >= AAF::SAMPLES_PER_FRAME )	{
		samplesRead = SAMPLES_PER_FRAME;
		for	( int input = 0; input < _numInputs; input++ )	{
			size_t samplesReturned;
			samplesReturned = Previous(input)->GetSamples( _inputBuffers[input], SAMPLES_PER_FRAME );
			if	( samplesReturned < samplesRead )
				samplesRead = samplesReturned;
		}
		
		for	( int pos = 0; pos < samplesRead; pos++, samplesWritten++ )
			Mix( pos*channels, buffer + (samplesWritten*channels) );
			
		if	( samplesRead < SAMPLES_PER_FRAME )	{
			stop = true;
			break;
		}
	}
	
	if	( !stop and ((samplesWanted - samplesWritten) > 0) )	{
		size_t remaining = samplesWanted - samplesWritten;
		
		samplesRead = remaining;
		for	( int input = 0; input < _numInputs; input++ )	{
			size_t samplesReturned;
			samplesReturned = Previous(input)->GetSamples( _inputBuffers[input], remaining );
			if	( samplesReturned < samplesRead )
				samplesRead = samplesReturned;
		}
		
		for	( int pos = 0; pos < samplesRead; pos++, samplesWritten++ )
			Mix( pos, buffer + (samplesWritten*channels) );
	}
		
	return samplesWritten;	
}

}	// namespace closure
