/*	AbstractOscillator.cpp    */

/*
	The AM1 Synthesizer for Macintosh, version 0.50 beta
	Copyright (C) 2003 The University of Toledo.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

	1.    Redistributions of source code must retain the above copyright notice,
	      this list of conditions and the following disclaimer. 
	2.    Redistributions in binary form must reproduce the above copyright notice,
	      this list of conditions and the following disclaimer in the documentation
	      and/or other materials provided with the distribution. 
	3.    The names of the authors and of the University of Toledo may not be used
	      to endorse or promote products derived from this software without specific
	      prior written permission. 

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND THE UNIVERSITY OF TOLEDO "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR THE UNIVERSITY OF TOLEDO BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include	<cstdlib>
#include	<cmath>
#include	"audiomath.h"
#include	"AbstractOscillator.h"

using	namespace	std;
using	namespace	AnthonyAudioFramework;

namespace	AnthonyAudioClasses	{

// constructors

AbstractOscillator::AbstractOscillator() : AudioAbstract()
{
	_frequency = _phase = _curPhase = 0;
	_amplitude = 0.25;
	_pulseWidth = 0.5;
	InitMods();
}

AbstractOscillator::AbstractOscillator( AudioAbstract* out, size_t port ) : AudioAbstract( out, port )
{
	_frequency = _phase = _curPhase = 0;
	_amplitude = 0.25;
	_pulseWidth = 0.5;
	InitMods();
}

AbstractOscillator::AbstractOscillator( AudioAbstract* out, size_t port, double frequency ) : AudioAbstract( out, port )
{
	_frequency = frequency;
	_phase = _curPhase = 0;
	_amplitude = 0.25;
	_pulseWidth = 0.5;
	InitMods();
}

AbstractOscillator::AbstractOscillator( AudioAbstract* out, size_t port, double frequency, double amp ) : AudioAbstract( out, port )
{
	_frequency = frequency;
	_curPhase = _phase = 0;
	_amplitude = amp;
	_pulseWidth = 0.5;
	InitMods();
}

AbstractOscillator::AbstractOscillator( AudioAbstract* out, size_t port, double frequency, double amp, double pw ) : AudioAbstract( out, port )
{
	_frequency = frequency;
	_curPhase = _phase = 0;
	_amplitude = amp;
	_pulseWidth = pw;
	InitMods();
}


void	AbstractOscillator::InitMods()
{
	_freqModSrc	= NULL;
	_freqModAmt	= 0.0;
	_ampModSrc	= NULL;
	_ampModAmt	= 0.0;
	_pwModSrc	= NULL;
	_pwModAmt	= 0.0;
	
	return;
}


void	AbstractOscillator::ReadMods( double& freq, double& amp, double& pw )
{
	freq = _frequency;
	amp = _amplitude;
	
	if	(_freqModSrc)	freq = FrequencyTranspose( _frequency, (_freqModAmt * _freqModSrc->Read()) );
	else				freq = _frequency;
	if	(_ampModSrc)	amp = clip( (_amplitude + _ampModAmt * _ampModSrc->Read()), 0.0, 1.0 );
	else				amp = _amplitude;
	if	(_pwModSrc)		pw = clip( (_pulseWidth + _pwModAmt * _pwModSrc->Read()), 0.0, 1.0 );
	else				pw = _pulseWidth;
	
	return;
}

	
// oscillator sample control 

size_t AbstractOscillator::GetSamples( AudioSample* buffer, size_t samplesWanted )
{
	int		chan = Channels();
	double	freq, amp, pw;
	
	// calculate current oscillator parameters with modulation sources
	ReadMods( freq, amp, pw );
	
	// calculate the phase increment per sample
	double phaseIncr = freq / (double)SamplingRate();
	
	// return only an exact multiple of the # of channels of samples
	size_t	samplesToMake = samplesWanted - samplesWanted % chan;
	size_t	samplesMade = 0;
	AudioSample	value;
	
	for	( ; samplesMade < samplesToMake; samplesMade+= chan )	{
		value = (AudioSample)( GenerateSample( freq, amp, pw ) * amp );		// call virtual function to create sample
		for	( int ch = 0; ch < chan; ch++ )						// copy sample into buffer for each channel
			buffer[samplesMade+ch] = value;
			
		_curPhase = fmod( _curPhase + phaseIncr, 1.0 );					// update the oscillator's phase
	}
	
	return samplesMade;
}


void	AbstractOscillator::Reset()
{
	AudioAbstract::Reset();
	_curPhase = _phase;
	return;
}

		
void	AbstractOscillator::InitializeForPlay()
{
	AudioAbstract::InitializeForPlay();
	_curPhase = _phase;
	return;
}

		
}	// namespace closure
