/*	AbstractControlSource.h

	An abstract base class for audio control signals.  These signals are used as 
	modulation sources to audio-rate signals.  They have a separate "control rate"
	at which they change and can be read multiple times inbetween updates.
	Updates occur manually and the Update() method must be called at an appropriate
	time by the coder for each object individually.
	
	Anthony Kozar
	March 22, 2003
	
 */

/*
	The AM1 Synthesizer for Macintosh, version 0.50 beta
	Copyright (C) 2003 The University of Toledo.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

	1.    Redistributions of source code must retain the above copyright notice,
	      this list of conditions and the following disclaimer. 
	2.    Redistributions in binary form must reproduce the above copyright notice,
	      this list of conditions and the following disclaimer in the documentation
	      and/or other materials provided with the distribution. 
	3.    The names of the authors and of the University of Toledo may not be used
	      to endorse or promote products derived from this software without specific
	      prior written permission. 

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND THE UNIVERSITY OF TOLEDO "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR THE UNIVERSITY OF TOLEDO BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef	ANTHONY_ABSTRACT_CONTROL_SOURCE
#define	ANTHONY_ABSTRACT_CONTROL_SOURCE

namespace	AnthonyAudioFramework	{

class	AbstractControlSource
{
	public:
		AbstractControlSource( unsigned long controlRate );
		~AbstractControlSource();
		
		// Read() returns current value of this control source
		// values of all control sources should be between -1.0 and 1.0
		double		Read() const	{ return _value; }		

		// Update() tells CS to recalculate its value for the next period
		virtual void	Update();		

		// Reset(): control source reinits itself to time 0 settings
		virtual void	Reset();		

		// ControlRate() allows you to get/set the control rate (in cps)
		unsigned long	ControlRate() const			{ return _controlRate; }
		void			ControlRate( unsigned long rate )	{ _controlRate = rate; }
		
	protected:
		// Timer() returns the number of control periods elapsed
		unsigned long	Timer() const				{ return _timer; }
		void			Value( double v )				{ _value = v; }
		double		Value() const				{ return _value; }
		
	private:
		unsigned long	_controlRate;
		unsigned long	_timer;		// number of control periods elapsed
		double		_value;		// current value, -1.0 to 1.0

};


}	// namespace closure

#endif
