/*	AudioAbstract.cpp
	
	Based on software by and copyright Tim Kientzle, 1997.
	Changes and additions by Anthony M Kozar Jr.
*/

/*
	The AM1 Synthesizer for Macintosh, version 0.50 beta
	Copyright (C) 2003 The University of Toledo.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

	1.    Redistributions of source code must retain the above copyright notice,
	      this list of conditions and the following disclaimer. 
	2.    Redistributions in binary form must reproduce the above copyright notice,
	      this list of conditions and the following disclaimer in the documentation
	      and/or other materials provided with the distribution. 
	3.    The names of the authors and of the University of Toledo may not be used
	      to endorse or promote products derived from this software without specific
	      prior written permission. 

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND THE UNIVERSITY OF TOLEDO "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR THE UNIVERSITY OF TOLEDO BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
   Copyright 1997 Tim Kientzle.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
      This product includes software developed by Tim Kientzle
      and published in ``The Programmer's Guide to Sound.''
4. Neither the names of Tim Kientzle nor Addison-Wesley
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL TIM KIENTZLE OR ADDISON-WESLEY BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include	<algorithm>
#include	"AudioAbstract.h"

namespace	AnthonyAudioFramework	{

const	long	MIN_SAMPLING_RATE	= 8000;
const long	MAX_SAMPLING_RATE = 44100;
const long	PREFERRED_SAMPLING_RATE = 44100;

const	int	MIN_CHANNELS = 1;
const	int	MAX_CHANNELS = 10;
const	int	PREFERRED_CHANNELS = 1;


// constructors, destructors, and initialization routines

AudioAbstract::AudioAbstract()
{
	_previous = 0;
	_next = 0;
	_samplingRate = 0;  _samplingRateFrozen = false;
	_channels = 0;      _channelsFrozen = false;
}


AudioAbstract::AudioAbstract( AudioAbstract* audio_output, size_t port )
{
	_next = audio_output;
	_previous = 0;
	audio_output->Previous( this, port );
	_samplingRate = 0;  _samplingRateFrozen = false;
	_channels = 0;      _channelsFrozen = false;
}


AudioAbstract::~AudioAbstract()
{
	if	(_previous)	delete [] _previous;
}


void	AudioAbstract::Previous( AudioAbstract* a, size_t n )
{
	if	( !_previous )	AllocatePrevious();
	if	( n < PreviousCount() )	_previous[n] = a;
	else	{
		cerr	<< "An attempt was made to add an input to a non-existent port.\n";
		exit(1);
	}

	return;		
}


void	AudioAbstract::AllocatePrevious()
{
	if	( PreviousCount() > 0 )	{
		if	(_previous)	delete [] _previous;							// delete old _previous pointers
		_previous = new AudioAbstract* [ PreviousCount() ];					// and make new ones
	}
	if	( !_previous )	{										// exit on any error
		cerr	<< "Could not allocate AudioAbstract _previous pointers.\n";
		exit(1);
	}
	else	for	( int i = 0; i < PreviousCount(); i++ )	{					// otherwise, initialize pointers
		_previous[i] = NULL;
	}
	
	return;
}
				

void AudioAbstract::InitializeForPlay()
{
	// traverse the tree of AudioAbstract objects
	size_t	numInputs = PreviousCount();
	
	for	( int in = 0; in < numInputs; in++ )	{
		if	( Previous(in) )
			Previous(in)->InitializeForPlay();
		else	{
			cerr	<< "Playback failed because of a non-existent audio link!\n";	// abort if null input pointer
			exit(1);
		}
	}
	
	return;
}


void AudioAbstract::Reset()
{
	// traverse the tree of AudioAbstract objects
	size_t	numInputs = PreviousCount();
	
	for	( int in = 0; in < numInputs; in++ )	{
		if	( Previous(in) )
			Previous(in)->Reset();
		else	{
			cerr	<< "Reset() failed because of a non-existent audio link!\n";	// abort if null input pointer
			exit(1);
		}
	}
	
	return;
}


// sample-rate/channels negotiation logic

void AudioAbstract::NegotiateSamplingRate()
{
	if	( Next() ) 												// Are we the last?
		Next()->NegotiateSamplingRate();								// No, try next in chain
	else	{
		long	min = MIN_SAMPLING_RATE, 
			max = MAX_SAMPLING_RATE,
			preferred = PREFERRED_SAMPLING_RATE;
				
		MinMaxSamplingRate( &min, &max, &preferred );						// Get preferred values				
		if	( min > max )	{									// Check for ridiculous answers
			cerr << "Couldn't negotiate sampling rate.\n";
			exit(1);
		}
		SetSamplingRateRecursive(preferred);							// Set sample rate everywhere
	}
}


void AudioAbstract::MinMaxSamplingRate( long *min, long *max, long *preferred )
{
	size_t	numinputs = PreviousCount();
	long		minresult, maxresult, preferredresult,
			minfinal, maxfinal, preferredfinal;
	
	minfinal = *min; maxfinal = *max; preferredfinal = *preferred;				// don't change originals yet
	
	// negotiate with each chain of objects connected as inputs
	for	( size_t i = 0; i < numinputs; i++ )	{
		minresult = *min; maxresult = *max; preferredresult = *preferred;
		if	( Previous(i) )	{
			Previous(i)->MinMaxSamplingRate( &minresult, &maxresult, &preferredresult );
			// save only the maximum min & preferred and minimum max values
			if	( minresult > minfinal )			minfinal = minresult;
			if	( maxresult < maxfinal )			maxfinal = maxresult;
			if	( preferredresult > preferredfinal )	preferredfinal = preferredresult;
		}
	}
		
	*min = minfinal; *max = maxfinal; *preferred = preferredfinal;				// return values to caller
	if	( _samplingRate != 0 )	*preferred = _samplingRate;					// override with any explicit value
	if	( *preferred < *min )	*preferred = *min;						// common sense checks
	if	( *preferred > *max )	*preferred = *max;
	
	return;
}


void AudioAbstract::SetSamplingRateRecursive( long s )
{
	size_t	numinputs = PreviousCount();
	
	for	( size_t i = 0; i < numinputs; i++ )							// traverse the object "tree"
		if	( Previous(i) )	Previous(i)->SetSamplingRateRecursive(s);
		
	SamplingRate(s);												// Set it
	_samplingRateFrozen = true;										// Yes, we've negotiated
}


void AudioAbstract::NegotiateChannels()
{
	if	( Next() )												// find last in chain
		Next()->NegotiateChannels();
	else {
		int	min = MIN_CHANNELS,
			max = MAX_CHANNELS,
			preferred = PREFERRED_CHANNELS;
		
		MinMaxChannels( &min, &max, &preferred );
		if	( min > max )	{									// Check for ridiculous answers
			cerr << "Couldn't negotiate the number of channels.\n";
			exit(1);
		}
		
		SetChannelsRecursive(preferred);								// Set # channels everywhere
	}
}


void AudioAbstract::MinMaxChannels( int *min, int *max, int *preferred )
{
	size_t	numinputs = PreviousCount();
	int		minresult, maxresult, preferredresult,
			minfinal, maxfinal, preferredfinal;
	
	minfinal = *min; maxfinal = *max; preferredfinal = *preferred;				// don't change originals yet
	
	// negotiate with each chain of objects connected as inputs
	for	( size_t i = 0; i < numinputs; i++ )	{
		minresult = *min; maxresult = *max; preferredresult = *preferred;
		if	( Previous(i) )	{
			Previous(i)->MinMaxChannels( &minresult, &maxresult, &preferredresult );
			// save only the maximum min and minimum max & preferred values
			if	( minresult > minfinal )			minfinal = minresult;
			if	( maxresult < maxfinal )			maxfinal = maxresult;
			if	( preferredresult < preferredfinal )	preferredfinal = preferredresult;
		}
	}
		
	*min = minfinal; *max = maxfinal; *preferred = preferredfinal;				// return values to caller
	if	( _channels != 0 )	*preferred = _channels;						// override with any explicit value
	if	( *preferred < *min )	*preferred = *min;						// common sense checks
	if	( *preferred > *max )	*preferred = *max;
	
	return;
}


void	AudioAbstract::SetChannelsRecursive( int ch )
{
	size_t	numinputs = PreviousCount();
	
	for	( size_t i = 0; i < numinputs; i++ )							// traverse the object "tree"
		if	( Previous(i) )	Previous(i)->SetChannelsRecursive(ch);
	Channels(ch);
	_channelsFrozen = true;
}


// low-level byte moving methods

long ReadIntMsb(istream &in, int size) {
	if (size <= 0) return 0;
	long l = ReadIntMsb(in,size-1) << 8;
	l |= static_cast<long>(in.get()) & 255;
	return l;
}


long BytesToIntMsb(void *vBuff, int size) {
	unsigned char *buff = reinterpret_cast<unsigned char *>(vBuff);
	if (size <= 0) return 0;
	long l = BytesToIntMsb(buff,size-1) << 8;
	l |= static_cast<long>(buff[size-1]) & 255;
	return l;
}

long ReadIntLsb(istream &in, int size) {
	if (size <= 0) return 0;
	long l = static_cast<long>(in.get()) & 255;
	l |= ReadIntLsb(in,size-1)<<8;
	return l;
}

long BytesToIntLsb(void *vBuff, int size) {
	unsigned char *buff = reinterpret_cast<unsigned char *>(vBuff);
	if (size <= 0) return 0;
	long l = static_cast<long>(*buff) & 255;
	l |= BytesToIntLsb(buff+1,size-1)<<8;
	return l;
}

void SkipBytes(istream &in, int size) {
	while (size-- > 0)
	in.get();
}

void WriteIntMsb(ostream &out, long l, int size) {
	if (size <= 0) return;
	WriteIntMsb(out, l>>8, size-1); // Write MS Bytes
	out.put(l&255); // Write LS Byte
}

void WriteIntLsb(ostream &out, long l, int size) {
	if (size <= 0) return;
	out.put(l&255);  // Write LS Byte
	WriteIntLsb(out, l>>8, size-1); // Write rest
}

}	// namespace closure

