/*	AudioAbstract.h

	Abstract base class for modular audio synthesis and processing.
	
	Based on software by and copyright Tim Kientzle, 1997.	
	Changes and additions by Anthony M Kozar Jr.
	September 8, 2002
*/

/*
	The AM1 Synthesizer for Macintosh, version 0.50 beta
	Copyright (C) 2003 The University of Toledo.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

	1.    Redistributions of source code must retain the above copyright notice,
	      this list of conditions and the following disclaimer. 
	2.    Redistributions in binary form must reproduce the above copyright notice,
	      this list of conditions and the following disclaimer in the documentation
	      and/or other materials provided with the distribution. 
	3.    The names of the authors and of the University of Toledo may not be used
	      to endorse or promote products derived from this software without specific
	      prior written permission. 

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND THE UNIVERSITY OF TOLEDO "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR THE UNIVERSITY OF TOLEDO BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
   Copyright 1997 Tim Kientzle.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
      This product includes software developed by Tim Kientzle
      and published in ``The Programmer's Guide to Sound.''
4. Neither the names of Tim Kientzle nor Addison-Wesley
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL TIM KIENTZLE OR ADDISON-WESLEY BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef ANTHONY_AUDIO_ABSTRACT_H
#define ANTHONY_AUDIO_ABSTRACT_H

#include	<typeinfo>
#include	<iostream>
#include	<cstdlib>
#include	<cstddef>
#include	<climits>

using namespace std;

namespace	AnthonyAudioFramework	{

// This constant is a recommended number of samples to request 
// when objects call GetSamples().  Smaller values increase the 
// number of function calls while larger values may increase the
// memory requirements of objects.
const	size_t SAMPLES_PER_FRAME = 512;

typedef short AudioSample; 								// A single audio sample
typedef unsigned char AudioByte; 							// an 8-bit unsigned byte

const	AudioSample	AUDIOSAMPLE_MAX = SHRT_MAX;					// redefine these if you change
const AudioSample	AUDIOSAMPLE_MIN = SHRT_MIN;					// the AudioSample typedef

long	ReadIntMsb( std::istream& in, int bytes );
long	BytesToIntMsb( void* buff, int bytes );
long	ReadIntLsb( std::istream& in, int bytes );
long	BytesToIntLsb( void* buff, int bytes );
void	SkipBytes( std::istream& in, int bytes );
void	WriteIntMsb( std::ostream& out, long l, int bytes );
void	WriteIntLsb( std::ostream& out, long l, int bytes );

class AudioAbstract
{
	/* audio interface */
	
	private:
		AudioAbstract**	_previous;						// array of input objects
		AudioAbstract*	_next;						// the output object
	
	public:
		AudioAbstract();

		// WARNING!  I've changed this constructor drastically from
		// the original way it worked.  Instead of passing the previous 
		// AudioAbstract object in the chain, you now pass the _next_ 
		// object and the input number ("port") on that object you wish
		// to connect to.  This hopefully makes connecting things easier
		// now that AudioAbstract supports multiple inputs.
		AudioAbstract( AudioAbstract* audio_output, size_t port );
		
		virtual ~AudioAbstract();

		// You must overload this method to return the number of input
		// ports for your class.
		virtual size_t	PreviousCount() = 0;
		
		AudioAbstract*	Previous( size_t n );
		void			Previous( AudioAbstract* a, size_t n );
		AudioAbstract*	Next() 				{ return _next; }
		void			Next( AudioAbstract* a )	{ _next = a; }
	
		// Returns number of samples actually read, 0 on error.
		// This should always return the full request unless there is
		// an error or end-of-data.
		virtual size_t GetSamples( AudioSample* , size_t ) = 0;

		virtual size_t ReadBytes( AudioByte* buff, size_t length )	{
			//return Previous()->ReadBytes(buff,length);
			return (size_t)0;							// changed default behavior to fail
		}

		// Reset() is called when a note stops playing and then needs to 
		// restart playing again without completely reinitializing parameters
		// or reallocating memory & objects.
		// Override this method if your subclass needs to reset any internal
		// values before replaying a note.
		// You must call the superclass version of this method at the 
		// beginning of your code (it calls Reset() on all of the inputs).
		virtual void Reset();
		
	protected:
		// Call this method to have AudioAbstract allocate the 
		// Previous pointers.  It uses PreviousCount() to determine
		// how many to make.
		void	AllocatePrevious();
		
		// Override this method if your subclass needs to allocate
		// memory before playback (you cannot allocate it during playback).
		// Make sure to call the superclass version of this method
		// at the beginning of your code.
		virtual void InitializeForPlay();
		
	/* sample rate negotiation interface */
	
	private:
		long	_samplingRate;
		bool	_samplingRateFrozen;

	public:
		virtual long SamplingRate()	{
			if	(!_samplingRateFrozen)  				// Not frozen?
				NegotiateSamplingRate(); 				// Go figure it out
			return _samplingRate;
		}
		
		virtual void SamplingRate( long s )	{ 				// Set the sampling rate
			if	(_samplingRateFrozen)	{
				std::cerr	<< "Can't change sampling rate.\n";
				std::exit(1);
			}
			
			_samplingRate = s;
		}

		virtual void NegotiateSamplingRate();
		virtual void MinMaxSamplingRate( long *min, long *max, long *prefer );
		virtual void SetSamplingRateRecursive( long s );

	/* number of channels negotiation interface */

	private:
		long	_channels;
		bool	_channelsFrozen;
		
	public:
		virtual int Channels()	{
			if	(!_channelsFrozen)	NegotiateChannels();
			return _channels;
		}

		virtual void Channels( int ch )	{
			if	(_channelsFrozen)	{
				std::cerr	<< "Can't change number of channels.\n";
				std::exit(1);
			}
			_channels = ch;
		}

		virtual void NegotiateChannels();
		virtual void MinMaxChannels( int *min, int *max, int *preferred );
		virtual void SetChannelsRecursive( int s );
};


inline AudioAbstract* AudioAbstract::Previous( size_t n )
{
	if	( _previous and n < PreviousCount() )	return _previous[n];
	else	return NULL;
}


}	// namespace closure

#endif

