/*	AM1Instrument.cpp    */

/*
	The AM1 Synthesizer for Macintosh, version 0.50 beta
	Copyright (C) 2003 The University of Toledo.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

	1.    Redistributions of source code must retain the above copyright notice,
	      this list of conditions and the following disclaimer. 
	2.    Redistributions in binary form must reproduce the above copyright notice,
	      this list of conditions and the following disclaimer in the documentation
	      and/or other materials provided with the distribution. 
	3.    The names of the authors and of the University of Toledo may not be used
	      to endorse or promote products derived from this software without specific
	      prior written permission. 

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND THE UNIVERSITY OF TOLEDO "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR THE UNIVERSITY OF TOLEDO BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include	<sstream>
#include	<cstring>

#include	"AM1Instrument.h"
#include	"SineWave.h"
#include	"SquareWave.h"
#include	"SimpleMixer.h"
#include	"RingModulator.h"
#include	"ANDMixer.h"
#include	"ORMixer.h"
#include	"AmplitudeModulator.h"

#include	"NoteName.h"
#include	"gentables.h"
#include	"mystrlib.h"

using	namespace	std;
using	namespace	AnthonyAudioFramework;
using	namespace	AnthonyAudioClasses;


// constructor

AM1Instrument::AM1Instrument()
{
	algorithm = numOperators = envNumStages = envSustainStage = envReleaseStage = 0;
	polyphony = 0;
	overallAmp = 0.0;
	filter.create = false;
	
	for	( int i = 0; i < NUM_OPERATORS; i++)	{
		oscillators[i].type = sine;
		oscillators[i].courseTuning = 0.0;
		oscillators[i].fineTuning = 0.0;
		oscillators[i].centsPerSemitone = 0.0;
		oscillators[i].amplitude = 0.0;
		oscillators[i].pulsewidth = 0.0;
	}	
	for	( int i = 0; i < NUM_MIXERS; i++)	{
		mix_types[i] = mtAdd;
	}
	for	( int i = 0; i < NUM_WAVESHAPERS; i++)	{
		shapers[i].create = false;
		shapers[i].wetlevel = 0.0;
		shapers[i].drylevel = 0.0;
		shapers[i].table = NULL;
	}
	for	( int i = 0; i < NUM_AUX_ENVELOPES; i++)	{
		auxEnvelopes[i].startlevel = 0.0;
		auxEnvelopes[i].mode = lmOff;
		for	( int j = 0; j < NUM_ENV_STAGES; j++ )	{
			auxEnvelopes[i].time[j] = 0.0;
			auxEnvelopes[i].level[j] = 0.0;
		}
	}
	for	( int i = 0; i < NUM_LFOS; i++)	{
		LFOs[i].type = sine;
		LFOs[i].rate = LFOs[i].phase = 0.0;
	}
	for	( int i = 0; i < NUM_GLOBAL_LFOS; i++)	{
		globalLFOs[i].type = sine;
		globalLFOs[i].rate = globalLFOs[i].phase = 0.0;
	}
	for	( int i = 0; i < NUM_KEY_FOLLOWS; i++)	{
		keyFollows[i].centerKey = 0;
		keyFollows[i].amountBelow = 0.0;
		keyFollows[i].amountAbove = 0.0;
	}
}


// functions to read an instrument settings file

bool	AM1Instrument::ReadInstrumentFile( istream& fin )
{
	bool	foundEnd = false;											// haven't found keyword "end"
	bool	finishedOK = true;										// assume all OK until we find a problem
	
	while	( !foundEnd and finishedOK )	{
		stringstream	ss;
		string		line;
		string::size_type	pos;
		
		getline( fin, line );
		pos = line.find(';');										// search for a comment
		if	( pos != string::npos )	line = line.erase(pos);					// remove it if found
		if	( AnyNonWhitespace(line) )	{							// if line is not now blank
			ss.str(line);
			string	word;
			
#ifndef	DEBUG_INSTRUMENT
			cout	<< line << endl;
#endif

			ss >> word;
			word = strtolower(word);
			if	( word == "end" )	foundEnd = true;
			else	if	( word == "alg" )		ss >> algorithm;
			else	if	( word == "polyphony")	ss >> polyphony;
			else	if	( word == "op" )		finishedOK = ReadOpStatement(ss);
			else	if	( word == "vertex" )	finishedOK = ReadVertexStatement(ss);
			else	if	( word == "shape" )	finishedOK = ReadShapeStatement(ss,0);
			else	if	( word == "shape2" )	finishedOK = ReadShapeStatement(ss,1);
			else	if	( word == "filter" )	finishedOK = ReadFilterStatement(ss);
			else	if	( word == "amp" )		finishedOK = ReadAmpStatement(ss);
			else	if	( word == "env" )		finishedOK = ReadEnvStatement(ss);
			else	if	( word == "lfo" )		finishedOK = ReadLFOStatement(ss);
			else	if	( word == "glfo" )	finishedOK = ReadGLFOStatement(ss);
			else	if	( word == "key" )		finishedOK = ReadKeyStatement(ss);
			else	if	( word == "mod" )		finishedOK = ReadModStatement(ss);
			
		}
	}
	
	return finishedOK;
}

bool	AM1Instrument::ReadOpStatement( std::istream& sin )
{
	int		opnum;
	string	tuning;
	
	// read operator number, check bounds, and map to index oscillators array
	sin	>> opnum;
	if	( opnum < 1 or opnum > NUM_OPERATORS )	return false;
	--opnum;
	
	// read waveform type keyword
	if	( !ReadWaveformWord( sin, oscillators[opnum].type ) )		// not a valid waveform keyword
		return false;

	// read course tuning
	sin	>> tuning;
	if	( isalpha( tuning.at(0) ) )	{					// if the first character is a letter, then tuning should be a note name
		NoteName	nn(tuning);
		oscillators[opnum].courseTuning = nn.toFrequency();		// convert note name to frequency in Hertz
	}
	else	oscillators[opnum].courseTuning = atof( tuning.c_str() );	// otherwise, tuning should be a floating-point value
		
	// read other tuning values and amplitude
	sin	>> oscillators[opnum].fineTuning
		>> oscillators[opnum].centsPerSemitone
		>> oscillators[opnum].amplitude;
		
	// read pulsewidth value and scale it (only if waveform is square)
	if	( oscillators[opnum].type == sqr )	{
		sin	>> oscillators[opnum].pulsewidth;
		oscillators[opnum].pulsewidth /= 100.0;
	}
		
	// test for errors
	if	( sin.fail() )
		return false;
	else	return true;
}

bool	AM1Instrument::ReadVertexStatement( std::istream& sin )
{
	string nextword;
	
	for	( int i = 0; i < NUM_MIXERS; i++ )	{
		sin	>> nextword;
		if	( sin.fail() )	return true;				// it's OK it the statement doesn't have all NUM_MIXERS keywords
		if	( !IdentifyMixerTWord( nextword, mix_types[i] ) )	// but it must be a valid mixer_t keyword
			return false;
	}
	
	return true;
}

bool	AM1Instrument::ReadShapeStatement( std::istream& sin, int num )
{
	waveshape	type;
	double*	parameters;
	size_t	numParms;
	
	sin	>> shapers[num].wetlevel
		>> shapers[num].drylevel;
		
	if	( !ReadWaveshapeWord( sin, type ) )
		return false;
		
	// determine the number of parameters to the shape table function
	switch (type)	{
		case	wsSine:
			numParms = 0; break;
			
		case	wsHalfSine:
		case	wsHypTan:
		case	wsPD:
		case	wsSqr:
			numParms = 1; break;
			
		case	wsClip:
		case	wsZero:
			numParms = 2; break;
			
		case	wsCheb:
			sin	>> numParms; break;
	}
	
	// read the parameters
	if	( numParms > 0 )	{
		parameters = new	double[numParms];
		//if	(!parameters)	return false;
		for	(int i = 0; i < numParms; i++ )
			sin	>> parameters[i];
	}
	
	// check for errors
	if	( sin.fail() )	{
		if	(parameters)	delete [] parameters;
		return false;
	}
	
	// save pointer to waveshaping table and set create flag
	shapers[num].table = CreateTable( type, parameters, numParms );
	shapers[num].create = true;
	
	return true;
}

bool	AM1Instrument::ReadFilterStatement( std::istream& sin )
{
	// read the filter mode keyword
	if	( !ReadFilterModeWord( sin, filter.type ) )
		return false;
	
	// read the filter cutoff frequency and resonance
	sin	>> filter.cutoff
		>> filter.resonance;
		
	// test for errors
	if	( sin.fail() )	return false;
	
	// set flag to create filter and signal success
	filter.create = true;
	return true;
}

bool	AM1Instrument::ReadAmpStatement( std::istream& sin )
{
	// read the overall amplitude value
	sin	>> overallAmp;
	
	// read the envelope parameters
	sin	>> envNumStages;
	sin	>> envSustainStage;
	sin	>> envReleaseStage;
	for	( int stage = 0; stage < envNumStages; stage++ )	{
		pair<double,double> temp;
		sin >> temp.first;
		sin >> temp.second;
		envParms.push_back(temp);
	}
	
	// test for errors
	if	( sin.fail() )
		return false;
	else	return true;
}

bool	AM1Instrument::ReadEnvStatement( std::istream& sin )
{
	
	// test for errors
	if	( sin.fail() )
		return false;
	else	return true;
}

bool	AM1Instrument::ReadLFOStatement( std::istream& sin )
{
	int	lfonum;
	
	// read the LFO number, check bounds, map to array index
	sin	>> lfonum;
	if	( lfonum < 1 or lfonum > NUM_LFOS )	return false;
	--lfonum;
	
	// read waveform type keyword
	if	( !ReadWaveformWord( sin, LFOs[lfonum].type ) )			// not a valid waveform keyword
		return false;

	// read the frequency and phase
	sin	>> LFOs[lfonum].rate
		>> LFOs[lfonum].phase;
			
	return true;
}

bool	AM1Instrument::ReadGLFOStatement( std::istream& sin )
{
	int	lfonum;
	
	// read the LFO number, check bounds, map to array index
	sin	>> lfonum;
	if	( lfonum < 1 or lfonum > NUM_GLOBAL_LFOS )	return false;
	--lfonum;
	
	// read waveform type keyword
	if	( !ReadWaveformWord( sin, globalLFOs[lfonum].type ) )		// not a valid waveform keyword
		return false;

	// read the frequency and phase
	sin	>> globalLFOs[lfonum].rate
		>> globalLFOs[lfonum].phase;
			
	return true;
}

bool	AM1Instrument::ReadKeyStatement( std::istream& sin )
{
	string	centerNote;
	int		num;
	
	// read the key follow number, check bounds, map to array index
	sin	>> num;
	if	( num < 1 or num > NUM_KEY_FOLLOWS )	return false;
	--num;
	
	// read the center note
	sin	>> centerNote;
	if	( !sin.fail() )	{
		NoteName	nn(centerNote);
		keyFollows[num].centerKey = nn.toMidiNote();
	}
	
	// read the amounts
	sin	>> keyFollows[num].amountBelow
		>> keyFollows[num].amountAbove;
		
	// test for errors
	if	( sin.fail() )
		return false;
	else	return true;
}

bool	AM1Instrument::ReadModStatement( std::istream& sin )
{
	ModulationRouting	temp;
	
	// read source keyword
	if	( !ReadModSourceWord( sin, temp.source ) )	return false;
	
	// read source number if necessary
	if	( temp.source != msAttackVelocity and temp.source != msReleaseVelocity and temp.source != msAftertouch
		  and temp.source != msNoteNum and temp.source != msClock )	{
		sin	>> temp.sourceNum;
		--temp.sourceNum;
	}
	
	sin	>> temp.amount;
	
	// read the destination and dest-type keywords
	if	( !ReadModDestWord( sin, temp.dest ) )		return false;
	if	( !ReadDestTypeWord( sin, temp.destType ) )	return false;
		
	// read the destination number is necessary
	if	( temp.dest != mdShaper and temp.dest != mdFilter and temp.dest != mdAmp )	{
		sin	>> temp.destNum;
		--temp.destNum;
	}
		
	// test for errors
	if	( sin.fail() )	return false;
	
	modMatrix.push_back(temp);
	return true;
}

bool	AM1Instrument::ReadWaveformWord( std::istream& sin, waveform& type )
{
	string nextword;
	
	sin	>> nextword;
	
	if		( nextword == "sine" )	type = sine;
	else	if	( nextword == "sqr" )	type = sqr;
	else	if	( nextword == "saw" )	type = saw;
	else	if	( nextword == "tri" )	type = tri;
	else	if	( nextword == "rnd" )	type = rnd;
	else	if	( nextword == "sync" )	type = sync;
	else	return false;
	
	return true;
}

bool	AM1Instrument::IdentifyMixerTWord( std::string& word, mixer_t& type )
{
	if		( word == "add" )		type = mtAdd;
	else	if	( word == "ring" )	type = mtRing;
	else	if	( word == "and" )		type = mtAnd;
	else	if	( word == "or" )		type = mtOr;
	else	if	( word == "xor" )		type = mtXor;
	else	return false;
	
	return true;
}

bool	AM1Instrument::ReadWaveshapeWord( std::istream& sin, waveshape& type )
{
	string nextword;
	
	sin	>> nextword;
	
	if		( nextword == "sine" )		type = wsSine;
	else	if	( nextword == "halfsine" )	type = wsHalfSine;
	else	if	( nextword == "tanh" )		type = wsHypTan;
	else	if	( nextword == "clip" )		type = wsClip;
	else	if	( nextword == "zero" )		type = wsZero;
	else	if	( nextword == "pd" )		type = wsPD;
	else	if	( nextword == "sqr" )		type = wsSqr;
	else	if	( nextword == "cheb" )		type = wsCheb;
	else	return false;
	
	return true;
}

bool	AM1Instrument::ReadFilterModeWord( std::istream& sin, filtermode& type )
{
	string nextword;
	
	sin	>> nextword;
	
	if		( nextword == "low" )	type = fmLow;
	else	if	( nextword == "high" )	type = fmHigh;
	else	if	( nextword == "band" )	type = fmBand;
	else	if	( nextword == "notch" )	type = fmNotch;
	else	return false;
	
	return true;
}


bool	AM1Instrument::ReadModSourceWord( std::istream& sin, modsources& type )
{
	string nextword;
	
	sin	>> nextword;
	
	if		( nextword == "attv" )	type = msAttackVelocity;
	else	if	( nextword == "relv" )	type = msReleaseVelocity;
	else	if	( nextword == "cntl" )	type = msController;
	else	if	( nextword == "pres" )	type = msAftertouch;
	else	if	( nextword == "note" )	type = msNoteNum;
	else	if	( nextword == "env" )	type = msEnv;
	else	if	( nextword == "lfo" )	type = msLFO;
	else	if	( nextword == "glfo" )	type = msGLFO;
	else	if	( nextword == "rnd" )	type = msRnd;
	else	if	( nextword == "key" )	type = msKey;
	else	if	( nextword == "fun" )	type = msFun;
	else	if	( nextword == "clock" )	type = msClock;
	else	return false;
	
	return true;
}


bool	AM1Instrument::ReadModDestWord( std::istream& sin, moddests& type )
{
	string nextword;
	
	sin	>> nextword;
	
	if		( nextword == "op" )	type = mdOperator;
	else	if	( nextword == "shape" )	type = mdShaper;
	else	if	( nextword == "filt" )	type = mdFilter;
	else	if	( nextword == "amp" )	type = mdAmp;
	else	if	( nextword == "lfo" )	type = mdLFO;
	else	if	( nextword == "glfo" )	type = mdGLFO;
	else	if	( nextword == "env" )	type = mdEnv;
	else	if	( nextword == "fun" )	type = mdFun;
	else	return false;
	
	return true;
}


bool	AM1Instrument::ReadDestTypeWord( std::istream& sin, dest_t& type )
{
	string nextword;
	
	sin	>> nextword;
	
	if		( nextword == "amp" )		type = dtAmp;
	else	if	( nextword == "pwm" )		type = dtPWM;
	else	if	( nextword == "freq" )		type = dtFreq;
	else	if	( nextword == "wetlvl" )	type = dtWetLevel;
	else	if	( nextword == "drylvl" )	type = dtDryLevel;
	else	if	( nextword == "res" )		type = dtResonance;
	else	if	( nextword == "db" )		type = dtDecibels;
	else	if	( nextword == "att" )		type = dtAttack;
	else	if	( nextword == "rel" )		type = dtRelease;
	else	if	( nextword == "rate" )		type = dtRate;
	else	if	( nextword == "depth" )		type = dtDepth;
	else	if	( nextword == "trig" )		type = dtTrigger;
	else	if	( nextword == "moda" )		type = dtModA;
	else	if	( nextword == "modb" )		type = dtModB;
	else	return false;
	
	return true;
}


AudioSample* AM1Instrument::CreateTable( waveshape method, double parms[], size_t numParms )
{
	const	size_t	TABLE_SIZE = 65536;
	AudioSample*	theTable = new AudioSample[TABLE_SIZE];

	switch (method)	{
		case	wsSine:
			SineTable( theTable, TABLE_SIZE );
			break;		
		case	wsHalfSine:
			//HalfSineTable( theTable, TABLE_SIZE, parms[0] );
			break;
		case	wsHypTan:
			HyperbolicTanTable( theTable, TABLE_SIZE, parms[0] );
			break;
		case	wsClip:
			LinearClippingTable( theTable, TABLE_SIZE, parms[0], parms[1] );
			break;
		case	wsZero:
			//LinearZeroingTable( theTable, TABLE_SIZE, parms[0], parms[1] );
			break;	
		case	wsPD:
			//PhaseDistortionTable( theTable, TABLE_SIZE, parms[0] );
			break;
		case	wsSqr:
			//InvertedSquareTable( theTable, TABLE_SIZE, parms[0] );
			break;		
		case	wsCheb:
			//ChebyshevPolynomialTable( theTable, TABLE_SIZE, parms, numParms );
			break;
	}
	
	return theTable;	
}
