/*	AM1Instrument.h

	A class which stores a set of configuration parameters for the 
	AM1 Synthesizer.  It can read text files which contain a description
	of those settings.
	
	Anthony Kozar
	July 23, 2003
 */

/*
	The AM1 Synthesizer for Macintosh, version 0.50 beta
	Copyright (C) 2003 The University of Toledo.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

	1.    Redistributions of source code must retain the above copyright notice,
	      this list of conditions and the following disclaimer. 
	2.    Redistributions in binary form must reproduce the above copyright notice,
	      this list of conditions and the following disclaimer in the documentation
	      and/or other materials provided with the distribution. 
	3.    The names of the authors and of the University of Toledo may not be used
	      to endorse or promote products derived from this software without specific
	      prior written permission. 

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND THE UNIVERSITY OF TOLEDO "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR THE UNIVERSITY OF TOLEDO BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef	ANTHONY_AM1_INSTRUMENT_H
#define	ANTHONY_AM1_INSTRUMENT_H

#include	<iostream>
#include	<vector>
#include	<utility>
#include	<cstdlib>

#include	"AudioAbstract.h"
#include	"AbstractOscillator.h"
#include	"AbstractMixer.h"
#include	"midi.h"

namespace	AAF = AnthonyAudioFramework;
namespace	AAC = AnthonyAudioClasses;
namespace	AM = AnthonyMidi;


const int	NUM_OPERATORS = 6;
const int	NUM_MIXERS = 3;
const int	NUM_WAVESHAPERS = 2;
const int	NUM_AUX_ENVELOPES = 6;
const int	NUM_LFOS = 3;
const int	NUM_GLOBAL_LFOS = 2;
const int	NUM_KEY_FOLLOWS = 6;
const int	NUM_ENV_STAGES = 4;
		
enum	waveform	{ sine, sqr, saw, tri, rnd, sync };
enum	mixer_t	{ mtAdd, mtRing, mtAnd, mtOr, mtXor };
enum	loopmode	{ lmOff, lmForwards, lmBoth };
enum	waveshape	{ wsSine, wsHalfSine, wsHypTan, wsClip, wsZero, wsPD, wsSqr, wsCheb };
enum	filtermode	{ fmLow, fmHigh, fmBand, fmNotch };
enum	modsources	{ msAttackVelocity, msReleaseVelocity, msController, msAftertouch, msNoteNum, msEnv, msLFO, msGLFO, msRnd, msKey, msFun, msClock };
enum	moddests	{ mdOperator, mdShaper, mdFilter, mdAmp, mdLFO, mdGLFO, mdEnv, mdFun };
enum	dest_t	{ dtAmp, dtPWM, dtFreq, dtWetLevel, dtDryLevel, dtResonance, dtDecibels, dtAttack, dtRelease, dtRate, dtDepth, dtTrigger, dtModA, dtModB };

class	AM1Instrument
{
	public:
		struct OscillatorRecord
		{
			waveform	type;
			double	courseTuning;		// pitch in Hertz at middle C
			double	fineTuning;			// detuning amount from course tuning (in cents)
			double	centsPerSemitone;		// pitch interval between adjacent Midi note numbers
			double	amplitude;			// operator amplitude
			double	pulsewidth;			// duty cycle width for square oscillators
		};
		
		struct WaveshaperRecord
		{
			bool			create;
			double		wetlevel;
			double		drylevel;
			AAF::AudioSample*	table;
		};
		
		struct FilterRecord
		{
			bool		create;
			filtermode	type;
			double	cutoff;
			double	resonance;
		};
		
		struct EnvelopeRecord			// auxiliary envelopes
		{
			double	startlevel;
			double	time[NUM_ENV_STAGES];
			double	level[NUM_ENV_STAGES];
			loopmode	mode;
		};
		
		struct LFO_Record
		{
			waveform	type;
			double	rate;
			double	phase;
		};
		
		struct KeyFollowRecord
		{
			AM::byte	centerKey;
			double	amountBelow;
			double	amountAbove;
		};
		
		struct ModulationRouting
		{
			modsources	source;
			std::size_t	sourceNum;
			double	amount;
			moddests	dest;
			dest_t	destType;
			std::size_t	destNum;
		};
		
	public:
		AM1Instrument();
				
		int	algorithm, numOperators, envNumStages, envSustainStage, envReleaseStage;
		std::size_t	polyphony;
		double overallAmp;
		
		// audio component settings
		OscillatorRecord		oscillators[NUM_OPERATORS];
		mixer_t			mix_types[NUM_MIXERS];
		WaveshaperRecord		shapers[NUM_WAVESHAPERS];
		FilterRecord		filter;
		
		std::vector< std::pair<double,double> >	envParms;
		
		// modulation settings
		EnvelopeRecord		auxEnvelopes[NUM_AUX_ENVELOPES];
		LFO_Record			LFOs[NUM_LFOS];
		LFO_Record			globalLFOs[NUM_GLOBAL_LFOS];
		KeyFollowRecord		keyFollows[NUM_KEY_FOLLOWS];
		
		std::vector<ModulationRouting>		modMatrix;
		
		// public methods
		bool	ReadInstrumentFile( std::istream& fin );
		
	private:
		bool	ReadOpStatement( std::istream& sin );
		bool	ReadVertexStatement( std::istream& sin );
		bool	ReadShapeStatement( std::istream& sin, int num );
		bool	ReadFilterStatement( std::istream& sin );
		bool	ReadAmpStatement( std::istream& sin );
		bool	ReadEnvStatement( std::istream& sin );
		bool	ReadLFOStatement( std::istream& sin );
		bool	ReadGLFOStatement( std::istream& sin );
		bool	ReadKeyStatement( std::istream& sin );
		bool	ReadModStatement( std::istream& sin );
		bool	ReadWaveformWord( std::istream& sin, waveform& type );
		bool	IdentifyMixerTWord( std::string& word, mixer_t& type );
		bool	ReadWaveshapeWord( std::istream& sin, waveshape& type );
		bool	ReadFilterModeWord( std::istream& sin, filtermode& type );
		bool	ReadModSourceWord( std::istream& sin, modsources& type );
		bool	ReadModDestWord( std::istream& sin, moddests& type );
		bool	ReadDestTypeWord( std::istream& sin, dest_t& type );
		
		AAF::AudioSample* CreateTable( waveshape method, double parms[], std::size_t numParms );
		
};

#endif

