/*	AM1 Synthesizer

	A software synthesizer combining square and sine waves using
	amplitude modulation and logic modulation.  Provides simple
	waveshaping, filtering, and multi-tap delay DSP.

	Note:  This file contains a small amount of code from the OMS 
	sample program and from Tim Kientzle's "A Programmer's Guide to Sound."
	
	Anthony Kozar
	December 5, 2002 (formerly Square Melody)
	March 23, 2003
 */

/*
	The AM1 Synthesizer for Macintosh, version 0.50 beta
	Copyright (C) 2003 The University of Toledo.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

	1.    Redistributions of source code must retain the above copyright notice,
	      this list of conditions and the following disclaimer. 
	2.    Redistributions in binary form must reproduce the above copyright notice,
	      this list of conditions and the following disclaimer in the documentation
	      and/or other materials provided with the distribution. 
	3.    The names of the authors and of the University of Toledo may not be used
	      to endorse or promote products derived from this software without specific
	      prior written permission. 

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND THE UNIVERSITY OF TOLEDO "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR THE UNIVERSITY OF TOLEDO BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include	<iostream>
#include	<fstream>
#include	<string>
#include	<vector>
#include	<queue>
#include	<utility>
#include	<cstdlib>
#include	<cassert>

#include	"MacPlayer.h"
#include	"AudioGate.h"
#include	"AmpEnvRelease.h"
#include	"Waveshaper.h"
#include	"SimpleMixer.h"
#include	"AmplitudeModulator.h"
#include	"SquareWave.h"
#include	"SineWave.h"
#include	"CSMidiController.h"
#include	"CSLFO.h"
#include	"CSLFOSine.h"
#include	"CSLFOTriangle.h"
#include	"CSLFOSquare.h"
#include	"CSMult.h"

#include	"macinit.h"
#include	"macevent.h"
#include	"macerror.h"
#include	"resources.h"
#include	"midi.h"
#include	"OMS.h"
#include	"OMSDeviceMenu.h"
#include	"NoteName.h"
#include	"gentables.h"

#include	"AM1Instrument.h"
#include	"AM1Note.h"
#include	"AM1NoteFactory.h"


using namespace	std;
using namespace	AnthonyAudioFramework;
using	namespace	AnthonyAudioClasses;
using	namespace	AnthonyMidi;

/*	Global Declarations    */

const	string			DEFAULT_INSTRUMENT("default.synth");
const	OSType			AM1_SIGNATURE = 'AM1s',
					OMS_INPORT_TYPE = 'in  ';
const	size_t			WSTABLE_SIZE = 65536;
					
const	double			SQUARE_AMP = 0.5;
const	size_t			DEFAULT_POLYPHONY = 4;

AM1Instrument*			gCurrentInstrument;
AM1Note**				gNotes;
size_t				gPolyphony = 4;
//size_t				gNextNote = 0;
pair<byte,unsigned long>*	gNoteOns;				// byte for midinote and long for timestamp

MacPlayer*				gPlayer;
SimpleMixer*			gNoteMixer;
AudioGate**				gGates;

const	size_t			MIDI_BUFFER_SIZE = 40;
OMSPacket				gMidiBuffer[MIDI_BUFFER_SIZE];
size_t				gBufferLastRead = 0, 
					gBufferLastWritten = 0;

ChannelStatus			gMidiOmniStatus;
int					gMidiReceiveChannel = -1;	// -1 means "any channel," 0-15 mean channels 1-16

short	gInputPortRefNum;							// refNum of the OMS input port
short	gCompatMode = omsModeAllowNonOMSApps;			// current OMS compatibility mode
short gChosenInputID = 0;						// uniqueID of selected input; 0 means none
short	gOutNodeRefNum = -1;						// node refNum of the selected output; -1 means non-existant
OMSDeviceMenuH	gInputMenu;

WindowPtr	gMidiSettingsDialog = NULL;


/*	Function Prototypes    */

OSErr	InitOMS( OSType appsig, OSType inport );
bool	LoadInstrument( const string& filename /*, vector< pair<double,double> >& envParms, size_t sustain, size_t release*/ );
void	DrawMidiSettingsDialog( WindowPtr win );
void	UpdateControlSources();
void	InitNote();
void	DeleteNote();
void	PlayNote( byte noteNum, byte attVel );
void	StopNote( byte noteNum, byte relVel );

#include	"macmain.h"
#include	"midifunc.h"

/*	Main Program Starts    */

int	main()
{	
	OSErr	error;
	
	MacInit::InitToolbox();	
	MacInit::InitSIOUX("\pAM1 Synthesizer");								// console window title
	MacInit::InitApplicationMenus();

	// !!!! Should test here to make sure that needed
	// !!!! Sound Manager features are available.
	
	error = InitOMS( AM1_SIGNATURE, OMS_INPORT_TYPE );						// initialize OMS
	if	( error != noErr )	{
		DisplayStopAlert( DEFAULT_ALERT, "Unable to initialize OMS.", "", "OMS must be installed for AM1 Synthesizer to run.", "" );
		return EXIT_FAILURE;
	}
	
	BuildWindow();
	
	gPlayer = NULL;
	//DoOpenInstrumentItem();
	if	( !LoadInstrument(DEFAULT_INSTRUMENT) )	DoQuit();					// FIX ME: shouldn't quit
	
	// This code loops because to change instruments, gPlayer has to be deleted.
	// But this cannot be done while gPlayer is in its Play() loop.  The Open
	// Instrument... menu item gets called from within the event loop which runs
	// inside of gPlayer's Play() loop, however.  So Open Instrument... just stops
	// gPlayer which eventually returns back to this point, and the note data
	// structures are recreated and we call gPlayer->Play() again.
	
	while (true)	{
		// start playing silent note
		InitNote();
		gPlayer->Play( MacEvent::CheckEvent, UpdateControlSources );			// event loop called from gPlayer
		DeleteNote();
	}
	
	//MacEvent::EventLoop();
	DoQuit();
	return EXIT_SUCCESS;
}
		

OSErr	InitOMS( OSType appsig, OSType inport )
{
	OSErr	err;
	OMSAppHookUPP appHook;
	OMSReadHookUPP readHook;
	
	if	( OMSVersion() == 0 )	return 1;
	
	appHook = NewOMSAppHook(MyAppHook);
	readHook = NewOMSReadHook(MyReadHook);
	
	err = OMSSignIn(appsig, SetCurrentA5(), LMGetCurApName(), appHook, &gCompatMode);
	//err = OMSSignIn(appsig, (long)LMGetCurrentA5(), LMGetCurApName(), appHook, &gCompatMode);
	/*	Passing CurrentA5 as the refCon solves the problem of A5 setup in the appHook.
		Using other Apple-recommended techniques for setting up A5 in the appHook
		are fine as well.  The client name will be the same as the application's name,
		as stored in the low-memory global CurApName. */
	
	if (err)
		return err;
	
	// Add an input port
	err = OMSAddPort(appsig, inport, omsPortTypeInput, (OMSReadHook2UPP)readHook, SetCurrentA5(), &gInputPortRefNum);
	if	(err)	{
		OMSSignOut(appsig);
		return err;
	}
	
	return noErr;
}

bool	LoadInstrument( const string& filename /*, vector< pair<double,double> >& envParms, size_t sustain, size_t release*/ )
{	
	/*int		numStages;
	int		numOsc;
	double	tempo;
	double	frequency;
	double	dur;*/
	ifstream	fin;
		
	// open input file
	fin.open( filename.c_str() );
	if	(!fin)	{
		DisplayStopAlert( DEFAULT_ALERT, "Unable to open the instrument settings file", filename.c_str(), "", "" );
		return false;
	}
	
	// read file into temp
	AM1Instrument*	tempInstr;
	
	tempInstr = new AM1Instrument();
	if	(tempInstr)	{
		if	( !tempInstr->ReadInstrumentFile(fin) )	{
			DisplayStopAlert( DEFAULT_ALERT, "There was an error reading the instrument settings file", filename.c_str(), "", "" );
			delete tempInstr;
			fin.close();
			return false;
		}
	}
	else	{
		DisplayStopAlert( DEFAULT_ALERT, "Out of memory! Unable to complete the load operation.", "", "", "" );
		fin.close();
		return false;
	}
	
	// if successful, replace any resident instrument settings with new ones
	if	(gCurrentInstrument)	delete gCurrentInstrument;
	gCurrentInstrument = tempInstr;
	
	fin.close();
	return true;
}


pascal void MyAppHook(OMSAppHookMsg *pkt, long myRefCon) {}
pascal void	MyReadHook(OMSPacket *pkt, long myRefCon)
{
	/*	Set up A5 for access to globals during this routine. */
	long olda5 = SetA5(myRefCon);
	
	/*	Process the MIDI packet as efficiently as possible.  It is guaranteed to be MIDI
		data, not some high-level event.  The applications refCon (appRefCon) that was 
		passed to OMSOpenConnection is in the low order word of pkt->tStamp. 
		A convenient way for an application to determine the source of the MIDI data is for 
		it to pass a number identifying the source as the appRefCon to OMSOpenConnection.
		The high-order word of pkt->tStemp is the sources ioRefNum (not its uniqueID); 
		applications can also look at this to determine the source of the MIDI data. */
	
	/*	This sample program merely echoes input to the selected output */
	
	// only process packets on the current Midi Receive channel
	if	( gMidiReceiveChannel < 0 or (pkt->data[0] & maskMidiChannel) == (byte)gMidiReceiveChannel )	{
		// store messages in a buffer to be processed during next event loop
		gBufferLastWritten = (++gBufferLastWritten) % MIDI_BUFFER_SIZE;
		gMidiBuffer[gBufferLastWritten].data[0] = pkt->data[0];
		gMidiBuffer[gBufferLastWritten].data[1] = pkt->data[1];
		gMidiBuffer[gBufferLastWritten].data[2] = pkt->data[2];
		gMidiBuffer[gBufferLastWritten].data[3] = pkt->data[3];
	}
	
	SetA5(olda5);
}

void	OpenOrCloseConnection(Boolean opening) {
	OSErr err;
	OMSConnectionParams conn;
	
	if (gChosenInputID == 0)
		return;
	conn.nodeUniqueID = gChosenInputID;
	conn.appRefCon = 0;
	if (opening)
		err = OMSOpenConnections( AM1_SIGNATURE, OMS_INPORT_TYPE, 1, &conn, false );
	else OMSCloseConnections( AM1_SIGNATURE, OMS_INPORT_TYPE, 1, &conn );
}


void	InputMenuSideEffects(void) {
	OpenOrCloseConnection(false);
	gChosenInputID = (*gInputMenu)->select.uniqueID;
	OpenOrCloseConnection(true);
}

void	BuildWindow(void) {
	WindowPtr		win;
	WindowPeek		wpk;
	Rect			menuRect;
	ControlHandle	channelPopUp;
	
	win = GetNewWindow(MIDI_INPUT_DEVICE_WIND_ID, NULL, (WindowPtr)-1L);
	gMidiSettingsDialog = win;
	wpk = (WindowPeek)win;
	SetPort(win);	// NewOMSDeviceMenu attaches window to current port
	ShowWindow(win);

	// add pop-up menu labels
/*	MoveTo(10,20);
	DrawString( "\pInput device:"  );
	MoveTo(10,58);
	DrawString( "\pInput channel:"  );
*/	
	// add the menu for selecting the OMS input device
	SetRect(&menuRect, 16, 16, 156, 32);
	gInputMenu = NewOMSDeviceMenu(NULL, odmFrameBox, &menuRect, 
		omsIncludeInputs + omsIncludeReal + omsIncludeVirtual, NULL);
		
	// add the menu for selecting the input Midi channel
	channelPopUp = GetNewControl( CHANNEL_POPUP_CNTL_ID, win );
	wpk->refCon = (long)channelPopUp;						// save ControlHandle for later
	
	DrawMidiSettingsDialog(win);
	return;
}

void	DrawMidiSettingsDialog( WindowPtr win )
{
	GrafPort*	port;
	port = (GrafPort*)win;

	BeginUpdate(win);
	EraseRect(&port->portRect);

	// draw pop-up menu labels
	MoveTo(10,10);
	DrawString( "\pInput device:"  );
	MoveTo(10,48);
	DrawString( "\pInput channel:"  );

	// update the menus
	DrawOMSDeviceMenu(gInputMenu);
	UpdateControls( win, port->visRgn );

	EndUpdate(win);
	return;
}

void	DoCloseItem( WindowPtr win )
{
	if	( win == gMidiSettingsDialog )	{
		HideWindow(win);
	}
	
	return;
}

void	DoAbout()
{
	DisplayStopAlert( DEFAULT_ALERT, "AM1 Synthesizer\r"
						   "By Anthony M. Kozar Jr.\r"
						   "August 2003\r\r"
						   "Copyright 2003 The University of Toledo\r\r"
						   "Portions Copyright 1997 Tim Kientzle.\r"
						   "All rights reserved.", "", "", "" );
	return;
}	

void	DoQuit(void) {
	OMSSignOut(AM1_SIGNATURE);
	ExitToShell();
}

void	DoOpenInstrumentItem()
{
	const	size_t	numTypes = 1;
	const	OSType	typeList[numTypes] = { 'TEXT' };
	string		filename;
	OSErr			err;
	short			fileref;
	 
	StandardFileReply	stdReply;

	StandardGetFile( NULL, numTypes, typeList, &stdReply );
	if (stdReply.sfGood)	{
		filename = reinterpret_cast<char *>(p2cstr(stdReply.sfFile.name));
		err = HSetVol( NULL, stdReply.sfFile.vRefNum, stdReply.sfFile.parID );		// set default directory
		
		if	(!LoadInstrument( filename /*, gEnvelope, gEnvSustain, gEnvRelease*/ ))	// load the instrument settings
			return;
		//DeleteNote();											// recreate audio classes
		//InitNote();
		if	(gPlayer)	gPlayer->Stop();								// by stopping current player
	}
	
	return;
}

void	DoSaveAudioToDiskItem()
{
	StandardFileReply	stdReply;

	StandardPutFile( "\pSave audio as:", "\pAM1.aiff", &stdReply );
	if (stdReply.sfGood)
		;	// do nothing for now

	return;
}

void	DoSetMidiInputDeviceItem()
{
	ShowWindow(gMidiSettingsDialog);
	SelectWindow(gMidiSettingsDialog);
	return;
}

void	DoAllNotesOff()
{
	for	( int i = 0; i < gPolyphony; i++ )	{
		gNotes[i]->NoteOff(0);
		gGates[i]->Gate( AudioGate::closed );
	}
	
	return;

}

bool	DoUpdate( EventRecord& evt )
{
	bool		evtHandled = false;
	WindowPtr	whichWin;
	GrafPtr	oldPort;
	
	GetPort(&oldPort);	
	whichWin = (WindowPtr)evt.message;
	SetPort(whichWin);
	
	if	(  whichWin == gMidiSettingsDialog )	{
		DrawMidiSettingsDialog(whichWin);
		evtHandled = true;
	}
	
	SetPort(oldPort);
	return evtHandled;
}

bool	DoWindowClick( EventRecord& evt, WindowPtr win )
{
	Point			pt;
	ControlHandle	whichControl;
	long			item;
	GrafPtr		oldPort;
	
	GetPort(&oldPort);	
	pt = evt.where;
	SetPort(win);
	GlobalToLocal(&pt);
	
	if	( win == gMidiSettingsDialog )	{
		// handle click in the OMS device menu
		if (TestOMSDeviceMenu(gInputMenu, pt)) {
			if (ClickOMSDeviceMenu(gInputMenu)) {
				InputMenuSideEffects();
			}
		}
		// or handle click in the Midi channel menu
		else {
			FindControl( pt, win, &whichControl );
			if	( whichControl == (ControlHandle)((WindowPeek)gMidiSettingsDialog)->refCon )	{
				TrackControl( whichControl, pt, (RoutineDescriptor*)-1 );
				item = GetControlValue(whichControl);
				gMidiReceiveChannel = item - 2;	// this maps choices 1-16 to 0-15 and "Any" to -1
			}
		}
	}
	
	SetPort(oldPort);
	return false;
}

void	DoResizeWindow( WindowPtr win )
{
/*	GrafPtr	savedPort;
	
	GetPort(&savedPort);
	SetPort(win);
	EraseRect( &(win->portRect) );
	
	if	(  win == gMidiSettingsDialog )	DrawMidiSettingsDialog(win);
	
	SetPort(savedPort);
*/	return;
}

void	ProcessMidiPackets()
{
	OMSPacket*	pkt;
	
	while	( gBufferLastRead != gBufferLastWritten )	{
		gBufferLastRead = (++gBufferLastRead) % MIDI_BUFFER_SIZE;
		pkt = &gMidiBuffer[gBufferLastRead];
		byte message = (pkt->data[0] & maskMidiStatusMessage);
		switch (message)	{
			case	kMidiNoteOn:
				//cout	<< "Note ON:\tNote Num " << pkt->data[1] << "\tVelocity " << pkt->data[2] << endl;
				if	( pkt->data[2] != 0 )
					PlayNote( pkt->data[1], pkt->data[2] );			// Midi NoteNum and Attack Velocity
				else	StopNote( pkt->data[1], pkt->data[2] );			// Vel. of 0 is a Note Off
				break;
				
			case	kMidiNoteOff:
				//cout	<< "Note OFF:\tNote Num " << pkt->data[1] << "\tRelease Velocity " << pkt->data[2] << endl;
				StopNote( pkt->data[1], pkt->data[2] );				// Midi NoteNum and Release Velocity
				break;
			
			case	kMidiPolyAftertouch:
				// ignore
				break;
					
			case	kMidiController:
				//cout	<< "Controller:\tController Num " << pkt->data[1] << "\tValue " << pkt->data[2] << endl;
				gMidiOmniStatus.controller[pkt->data[1]] = pkt->data[2];	// Controller number and value
				break;
					
			case	kMidiProgramChange:
				//cout	<< "Program Change:\tProgram Num " << pkt->data[1] << endl;
				gMidiOmniStatus.activeProgram = pkt->data[1];			// Program number
				break;
					
			case	kMidiChannelPressure:
				//cout	<< "Channel Pressure:\tValue " << pkt->data[1] << endl;
				gMidiOmniStatus.monoPressure = pkt->data[1];			// channel pressure value
				break;
					
			case	kMidiPitchBend:
				//cout	<< "Pitch Bend:\tlow " << pkt->data[1] << "\thigh " << pkt->data[2] << endl;
				gMidiOmniStatus.pitchbend = Make14bitInteger( pkt->data[1], pkt->data[2] );
				break;
				
			default:
				//cout	<< "Midi Packet: " << pkt->data[0] << " " << pkt->data[1] << " " 
				//	<< pkt->data[2] << " " << pkt->data[3] << " " << endl;
				break;
		}
	}
	
	return;
}

byte	ReadController( byte chan, byte con )
{
	return gMidiOmniStatus.controller[con];
}

byte	ReadAftertouch( byte chan )
{
	return gMidiOmniStatus.monoPressure;
}

void	UpdateControlSources()
{
	for	( int i = 0; i < gPolyphony; i++ )
		gNotes[i]->UpdateControls();
	return;
}

void	InitNote()
{
	AM1NoteFactory	factory(gCurrentInstrument);
	
	if	(gCurrentInstrument->polyphony > 0)
		gPolyphony = gCurrentInstrument->polyphony;
	else	gPolyphony = DEFAULT_POLYPHONY;
		
	#ifndef	DEBUG_INSTRUMENT
	cout	<< endl << "polyphony: " << gPolyphony << endl << endl;
	#endif

	gPlayer = new MacPlayer();
	gNoteMixer = new SimpleMixer( gPlayer, 0, gPolyphony );
	gGates = new AudioGate*[gPolyphony];
	gNotes = new AM1Note*[gPolyphony];
	gNoteOns = new pair<byte,unsigned long>[gPolyphony];
	
	for	( int i = 0; i < gPolyphony; i++ )	{
		gGates[i] = new AudioGate( gNoteMixer, i, AudioGate::closed );
		gNotes[i] = factory.BuildNote( gGates[i], 0 );
		gNoteOns[i].first = 255; gNoteOns[i].second = 0;	// 255 = "note inactive"
	}
		
	return;		
}

void	DeleteNote()
{
	for	( int i = 0; i < gPolyphony; i++ )	{
		if	(gGates[i])	delete gGates[i];
		if	(gNotes[i])	delete gNotes[i];
	}
	
	if	(gNotes)		{ delete gNotes; gNotes = NULL; }
	if	(gGates)		{ delete gGates; gGates = NULL; }
	if	(gPlayer)		{ delete gPlayer; gPlayer = NULL; }
	if	(gNoteOns)		{ delete gNoteOns; gNoteOns = NULL; }
	return;
}

void	PlayNote( byte noteNum, byte attVel )
{	
	static unsigned long timestamp = 0;					// counter for marking "time" notes start
	
	// find an available note
	unsigned long	tsMin = ULONG_MAX;
	int			nextNote = -1, minNote = -1;
	for	(int i = 0; i < gPolyphony; i++ )	{
		if	( gNoteOns[i].first == 255 )	{			// 255 = note inactive
			nextNote = i;
			break;
		}
		else	if	( gNoteOns[i].second <= tsMin )	{
			tsMin = gNoteOns[i].second;				// remember the minimum timestamp
			minNote = i;
		}
	}
	if	( nextNote < 0 )	nextNote = minNote;			// steal the oldest note if none are available
	
	// reinitialize the note
	gNotes[nextNote]->NoteOn( noteNum, attVel);
	gNotes[nextNote]->ResetControls();		
	gGates[nextNote]->Reset();
	
	gNoteOns[nextNote].first = noteNum;					// save which midi notes are active
	gNoteOns[nextNote].second = ++timestamp;				// save and increment the timestamp
	
	return;
}

void	StopNote( byte noteNum, byte relVel )
{
	// find which gNotes corresponds to the NoteOff message and turn it off
	for	( int i = 0; i < gPolyphony; i++ )	{
		if	( gNoteOns[i].first == noteNum )	{
			gNotes[i]->NoteOff(relVel);
			gNoteOns[i].first = 255;				// mark note as inactive
		}
		/*else	if	( gNoteOns[i].first == 255 )
			gGates[i]->Gate( AudioGate::closed );*/		// gate previously inactive notes
	}
	
	return;
}
